/*
 * eth_ecat_custom_tuple.hpp
 *
 *  Created on: May 5, 2023
 *      Author: algin
 */

#ifndef FREE_RTOS_ETHERNET_INDUSTRY_ETH_ECAT_CUSTOM_TUPLE_HPP_
#define FREE_RTOS_ETHERNET_INDUSTRY_ETH_ECAT_CUSTOM_TUPLE_HPP_

namespace free_rtos {

// Базовый шаблон класса, никогда не инстанциируется, поэтому без тела
template<typename... Args>
struct custom_tuple;

// Основная специализация шаблона. Есть еще одна для протокола CoE.
template<typename HeadT, typename... TailT>
struct custom_tuple<HeadT, TailT...> : custom_tuple<TailT...> {
    custom_tuple(HeadT head, TailT... tail)
    : custom_tuple<TailT...>(tail...)
    , head_(head) { }

    custom_tuple()
    : custom_tuple<TailT...>() { }

    using TBase = custom_tuple<TailT...>;
    using THead = HeadT;
    using THeadDeref = typename std::remove_reference<THead>::type;

    constexpr static size_t size = sizeof(THeadDeref) + TBase::size;

    THead head_;

    uint8_t* pack(uint8_t *raw) {
        THeadDeref *head = new(raw) THeadDeref{head_};

        (void)head;

        return TBase::pack(raw + sizeof(THeadDeref));
    }

    uint8_t* unpack(uint8_t *raw) {
        THeadDeref *head = reinterpret_cast<THeadDeref*>(raw);

        head_ = *head;

        return TBase::unpack(raw + sizeof(THeadDeref));
    }
};

// Специализация завершения рекурсии
template<>
struct custom_tuple<> {
    constexpr static size_t size = 0;

    uint8_t* pack(uint8_t *raw) {
        return raw;
    }

    uint8_t* unpack(uint8_t *raw) {
        return raw;
    }
};

template<size_t index, typename TupleT>
struct custom_tuple_element;

template<size_t index, typename HeadT, typename... TailT>
struct custom_tuple_element<index, custom_tuple<HeadT, TailT...>> {
    using TBase = custom_tuple_element<index - 1, custom_tuple<TailT...>>;
    using type = typename TBase::type;

    static type& get(custom_tuple<HeadT, TailT...>& t) {
        return TBase::get(t);
    }
};

template<typename HeadT, typename... TailT>
struct custom_tuple_element<0, custom_tuple<HeadT, TailT...>> {
    using type = HeadT;

    static type& get(custom_tuple<HeadT, TailT...>& t) {
        return t.head_;
    }
};

template<size_t index, typename TupleT>
typename custom_tuple_element<index, TupleT>::type& get(TupleT& t) {
    return custom_tuple_element<index, TupleT>::get(t);
}

}

#endif /* FREE_RTOS_ETHERNET_INDUSTRY_ETH_ECAT_CUSTOM_TUPLE_HPP_ */
